@@ -1,34 +1,18 @@
 from abc import ABC, abstractmethod
-from collections import Iterable
+from collections.abc import Iterable
 from copy import copy
 from itertools import product as iter_product
 
 import numpy as np
 from scipy.sparse import issparse, lil_matrix
 
-from kernelmethods.operations import center_km, is_PSD, normalize_km, normalize_km_2sample
+from kernelmethods.operations import center_km, is_PSD, normalize_km, \
+    normalize_km_2sample, frobenius_norm
 from kernelmethods.utils import check_callable, ensure_ndarray_1D, ensure_ndarray_2D, \
-    get_callable_name, not_symmetric
+    get_callable_name, not_symmetric, contains_nan_inf
 from kernelmethods import config as cfg
-
-class KernelMethodsException(Exception):
-    """
-    Generic exception to indicate invalid use of the kernelmethods library.
-
-
-    Allows to distinguish improper use of KernelMatrix from other code exceptions
-    """
-    pass
-
-
-class KMAccessError(KernelMethodsException):
-    """Exception to indicate invalid access to the kernel matrix!"""
-    pass
-
-
-class KMSetAdditionError(KernelMethodsException):
-    """Exception to indicate invalid addition of kernel matrix to a KernelSet"""
-    pass
+from kernelmethods.config import KernelMethodsException, KMAccessError, \
+    KMSetAdditionError
 
 
 class BaseKernelFunction(ABC):
@@ -285,6 +269,11 @@ class KernelMatrix(object):
     def attributes(self):
         """
         Returns all the attributes currently set.
+
+        Returns
+        -------
+        attributes : dict
+            Dict of the all the attributes currently set.
         """
 
         return self._attr
@@ -357,6 +346,10 @@ class KernelMatrix(object):
         if not self._populated_fully:
             self._populate_fully(fill_lower_tri=True, dense_fmt=True)
 
+        if contains_nan_inf(self._full_km):
+            raise Warning('Kernel matrix computation resulted in Inf or NaN '
+                          'values - check your parameters and data!')
+
         if self._keep_normed:
             if not self._is_normed:
                 self.normalize()
@@ -369,11 +362,23 @@ class KernelMatrix(object):
     def full_sparse(self):
         """Kernel matrix populated in upper tri in sparse array format."""
 
-        return self._populate_fully(fill_lower_tri=False)
+        return self._populate_fully(dense_fmt=False, fill_lower_tri=False)
 
 
     def center(self):
-        """Method to center the kernel matrix"""
+        """
+        Method to center the kernel matrix
+
+        Returns
+        -------
+        None
+
+        Raises
+        ------
+        NotImplementedError
+            If the KM is attached two separate samples.
+            Centering a KM is possible only when attached to a single sample.
+        """
 
         if self._two_samples:
             raise NotImplementedError('Centering is not implemented (or possible)'
@@ -388,11 +393,21 @@ class KernelMatrix(object):
 
     def normalize(self, method='cosine'):
         """
+
         Normalize the kernel matrix to have unit diagonal.
 
         Cosine normalization mplements definition according to Section 5.1 in
             Shawe-Taylor and Cristianini, "Kernels Methods for Pattern Analysis", 2004
 
+        Parameters
+        ----------
+        method : str
+            Identifier of the method.
+
+        Returns
+        -------
+        None
+
         """
 
         if not self._populated_fully:
@@ -417,6 +432,10 @@ class KernelMatrix(object):
                                                        KM_YY.diagonal())
             self._is_normed = True
 
+            if contains_nan_inf(self._normed_km):
+                raise Warning('normalization of kernel matrix resulted in Inf / NaN '
+                              'values - check your parameters and data!')
+
 
     @property
     def centered(self):
@@ -438,7 +457,8 @@ class KernelMatrix(object):
         if not self._populated_fully:
             self._populate_fully(dense_fmt=True, fill_lower_tri=True)
 
-        self._frob_norm = frobenius_norm(self._full_km)
+        if not hasattr(self, '_frob_norm'):
+            self._frob_norm = frobenius_norm(self._full_km)
 
         return self._frob_norm
 
@@ -640,6 +660,10 @@ class KernelMatrix(object):
         if issparse(self._full_km) and dense_fmt:
             self._full_km = self._full_km.todense()
 
+        if contains_nan_inf(self._full_km):
+            raise Warning('Kernel matrix computation resulted in Inf or NaN '
+                          'values - check your parameters and data!')
+
         return self._full_km
 
 
@@ -911,7 +935,8 @@ class KernelSet(object):
     def __getitem__(self, index):
         """To retrieve individual kernels"""
 
-        if not isinstance(index, int):
+        if not ( isinstance(index, int) or
+                 np.issubdtype(np.asanyarray(index).dtype, np.integer)):
             raise ValueError('Only integer indices are permitted, '
                              'accessing one KM at a time')
 
@@ -1014,7 +1039,9 @@ class KernelSet(object):
 
         for index in range(self.size):
             self._km_set[index].attach_to(sample, name_one=name)
-            if attr_name is not None:
+
+        if attr_name is not None:
+            for index in range(self.size):
                 self._km_set[index].set_attr(attr_name, attr_value)
 
 
